#include <math.h>

/*
 * Utilities
 */

static unsigned int
_mkid()
{
        static unsigned int id = 2;
        return id++;
}

static short
_min(short cost[])
{
        int i;
        short min = cost[1];
        for (i = 2; i <= _NTS; i++)
                if (cost[i] < min)
                        min = cost[i];
        return min;
}

/*
** Tables
*/

typedef struct _state *State;

typedef struct _state_key {
        int op;
        unsigned int id1;
        unsigned int id2;
} *StateKey;

typedef struct _swdc_key {
        int op;
        unsigned int id1;
        unsigned int id2;
        int (*dyncost)[_DRULES];
} *SwdcKey;

typedef union {
        StateKey state;
        SwdcKey  swdc;
        State    full;
} *Key;

typedef union {
        State state;
        int   id;
} *Val;

typedef struct _bucket {
        Key             key;
        Val             val;
        struct _bucket *next;
} *Bucket;

typedef struct _table {
        Bucket          t[_SIZE];
        unsigned int  (*hash)(const Key);
        bool          (*eq)(const Key, const Key);
} *BurmTable;

/*
** Hashing
*/

static unsigned int
hash_state(const Key key)
{
        StateKey k = key->state;
        unsigned int hashval = k->op * k->id1 * k->id2;
        return hashval % _SIZE;
}

static bool
eq_state(const Key key1, const Key key2)
{
        StateKey k1 = key1->state;
        StateKey k2 = key2->state;
        return (k1->op  == k2->op  &&
                k1->id1 == k2->id1 &&
                k1->id2 == k2->id2);
}

static unsigned int
hash_swdc(const Key key)
{
        SwdcKey k = key->swdc;
        int i;
        unsigned int hashval = k->op * k->id1 * k->id2;
        for (i = 0; i < _DRULES && (*(k->dyncost))[i] != -1; i++)
                hashval *= (*(k->dyncost))[i];
        return hashval % _SIZE;
}

static bool
eq_swdc(const Key key1, const Key key2)
{
        SwdcKey k1 = key1->swdc;
        SwdcKey k2 = key2->swdc;
        int i;
        if (k1->op  != k2->op ||
            k1->id1 != k2->id1 ||
            k1->id2 != k2->id2)
                return false;
        for (i = 0; i < _DRULES; i++)
                if ((*(k1->dyncost))[i] != (*(k2->dyncost))[i])
                        return false;
                else if ((*(k1->dyncost))[i] == -1)
                        break;
        return true;
}

static unsigned int
hash_full(const Key key)
{
        State s = key->full;
        int i;
        unsigned int hashval = 1;
        for (i = 1; i <= _NTS; i++) {
                hashval *= s->cost[i];
                hashval *= _rule(s, i);
        }
        return hashval % _SIZE;
}

static bool
eq_full(const Key key1, const Key key2)
{
        State s1 = key1->full;
        State s2 = key2->full;
        int i;
        if (!_ruleeq(s1, s2))
                return false;
        for (i = 1; i <= _NTS; i++)
                if (s1->cost[i] != s2->cost[i])
                        return false;
        return true;
}

/*
** Initialization
*/

static BurmTable state_cache; /* state_cache :: StateKey -> State */
static BurmTable swdc_cache;  /* swdc_cache  :: SwdcKey -> State  */
static BurmTable id_cache;    /* id_cache    :: State -> int      */

static void
_burminit()
{
        static bool flag = false;
        if (!flag) {
                flag              = true;
                state_cache       = (BurmTable) malloc(sizeof(*state_cache));
                swdc_cache        = (BurmTable) malloc(sizeof(*swdc_cache));
                id_cache          = (BurmTable) malloc(sizeof(*id_cache));
                state_cache->hash = hash_state;
                state_cache->eq   = eq_state;
                swdc_cache->hash  = hash_swdc;
                swdc_cache->eq    = eq_swdc;
                id_cache->hash    = hash_full;
                id_cache->eq      = eq_full;
        }
}

/*
** lookup/install
*/

static Val
_lookup(const BurmTable tab, const Key k)
{
        Bucket b;
        for (b = tab->t[tab->hash(k)]; b != NULL; b = b->next)
                if (tab->eq(k, b->key) == true)
                        return b->val;
        return NULL;
}

static void
_install(BurmTable tab, Key k, Val v)
{
        Bucket b;
        unsigned int hashval;
        if (_lookup(tab, k) == NULL) {
                b               = malloc(sizeof(*b));
                hashval         = tab->hash(k);
                b->key          = k;
                b->val          = v;
                b->next         = tab->t[hashval];
                tab->t[hashval] = b;
        } else
                fatal("_install", "OOPS\n", 0);
}

/*
** _label()
*/

// Stats
unsigned int _states_labeled = 0;
unsigned int _states_found   = 0;
unsigned int _swdcs_labeled  = 0;
unsigned int _swdcs_found    = 0;
unsigned int _unique_states  = 0;
unsigned int _ident_states   = 0;
unsigned int _indirs         = 0;
unsigned int _indir_reused   = 0;

static State _calc_label(NODEPTR_TYPE);

static void
_label(NODEPTR_TYPE a)
{
	if (!a)
		fatal("_label", "Null tree\n", 0);

        State p;
        int id1, id2, op = OP_LABEL(a);

        switch (_arity(op)) {
        case 0:
                id1 = id2 = 1;
                break;
        case 1:
                _label(LEFT_CHILD(a));
                id1 = STATE_ID(LEFT_CHILD(a));
                id2 = 1;
                break;
        case 2:
                _label(LEFT_CHILD(a));
                _label(RIGHT_CHILD(a));
                id1 = ((State) STATE_LABEL(LEFT_CHILD(a)))->id; // XXX
                id2 = ((State) STATE_LABEL(RIGHT_CHILD(a)))->id;
                break;
        default:
                fatal("_install", "OOPS2\n", 0);
        }

        _burminit();

        Key k = allocate(sizeof(*k), PERM);
        Val v;

	if (generic(op) == INDIR) {
                p = _calc_label(a);
                _indirs++;
        } else if (_is_tainted(op)) {
                k->swdc          = allocate(sizeof(*k->swdc), PERM);
                k->swdc->op      = op;
                k->swdc->id1     = id1;
                k->swdc->id2     = id2;
                k->swdc->dyncost = _dyncost(op, a);
                if ((v = _lookup(swdc_cache, k)) == NULL) {
                        p        = _calc_label(a);
                        v        = allocate(sizeof(*v), PERM);
                        v->state = p;
                        _install(swdc_cache, k, v);
                        _swdcs_labeled++;
                } else {
                        p = v->state;
                        _swdcs_found++;
                }
        } else {
                k->state      = allocate(sizeof(*k->state), PERM);
                k->state->op  = op;
                k->state->id1 = id1;
                k->state->id2 = id2;
                if ((v = _lookup(state_cache, k)) == NULL) {
                        p        = _calc_label(a);
                        v        = allocate(sizeof(*v), PERM);
                        v->state = p;
                        _install(state_cache, k, v);
                        _states_labeled++;
                } else {
                        p = v->state;
                        _states_found++;
                }
        }

        STATE_LABEL(a) = p;
}

static State
_calc_label(NODEPTR_TYPE a)
{
        // Former _label()
        State p = _label2(a);

        // Normalize
        int i, delta = _min(p->cost);
        for (i = 1; i <= _NTS; i++)
                if (!(p->cost[i] == 0x7fff))
                        p->cost[i] -= delta;

        // Set state number
        Key k   = allocate(sizeof(*k), PERM);
        k->full = p;
        Val v;

        if ((v = _lookup(id_cache, k)) == NULL) {
                int id = _mkid();
                v      = allocate(sizeof(*v), PERM);
                v->id  = id;
                p->id  = id;
                _install(id_cache, k, v);
                _unique_states++;
        } else {
                p->id = v->id;
                if (generic(a->op) == INDIR)
                        _indir_reused++;
                _ident_states++;
        }

        return p;
}

// eof
